/*
 * burger_potgieter.cc
 *
 *      Author: kappl
 */

#include "burgerPotgieter.h"
#include <cmath>
#include <chrono>
#include <iostream>

/**
 * @brief Constructor
 *
 * @param modelInfo Model information from import.cc
 * @param sp Particle properties
 */
BurgerPotgieter::BurgerPotgieter(const TmodelInformation &modelInfo, const Tproperties &sp) : Particle(modelInfo, sp)  {
	if (modelInfo.dt > 0)
	{
		dt = modelInfo.dt;
	}
	else
	{
		dt = 3000;
	}

	kappa0 = modelInfo.kappaScaling*20./9.*pow(10,-5);
	perPart = 0.05;
	heliosphereBoundary = 10.;
}

/**
 * @brief Destructor
 */
BurgerPotgieter::~BurgerPotgieter() {
}

/*
 * @brief Getter for parallel part of the diffusion tensor
 *
 * @return Parallel part of the diffusion tensor
 */
double BurgerPotgieter::getKappaParallel() const {
	return data.parallel;
}

/*
 * @brief Getter for radial part of the diffusion tensor
 *
 * @return Radial part of the diffusion tensor
 */
double BurgerPotgieter::getKappaRR() const {
	return getKappaParallel()/(1.+getTanPsi()*getTanPsi())
				+ getKappaPerpendicular()*getTanPsi()*getTanPsi()/(1.+getTanPsi()*getTanPsi());
}

/**
 * @brief Getter for drift in r direction
 *
 * @return radial drift
 */
double BurgerPotgieter::getDriftR() const {
	// Define Heaviside function depending term
	int heavy = getHeaviside();

	return getPolarity()*getCharge()*9./2.*pow(10,-14)*2./3.*getRigidity()*getBeta()/(Constants::A*BfieldScaling)*getR()*cos(getTheta())/sin(getTheta())*getTanPsi()/pow(1+getTanPsi()*getTanPsi(),2)*heavy;
}

/**
 * @brief Getter for drift in theta direction
 *
 * @return angular drift
 */
double BurgerPotgieter::getDriftTheta() const {
	// Define Heaviside function depending term
	int heavy = getHeaviside();

	return -getPolarity()*getCharge()*9./2.*pow(10,-14)*2./3.*getRigidity()*getBeta()/(Constants::A*BfieldScaling)*getR()*getTanPsi()*(2+getTanPsi()*getTanPsi())/pow(1+getTanPsi()*getTanPsi(),2)*heavy;
}

/**
 * @brief Getter for drift in the heliospheric current sheet (HCS) in radial direction
 *
 * @return HCS drift
 */
double BurgerPotgieter::getDriftSheetR() const {
	double deltaTheta = 9./4.*pow(10,-11)*2.*getRigidity()*getV()/(Constants::A*BfieldScaling*Constants::Omega*cos(convertDeg(getAngle())));

	// Look if the particle is inside the "cone" of the wavy heliospheric current sheet
	if (getTheta() > Constants::pi/2.-convertDeg(getAngle())-deltaTheta && getTheta() < Constants::pi/2.+convertDeg(getAngle())+deltaTheta)
	{
		return -getPolarity()*getCharge()*sin(getTheta())*getTanPsi()/sqrt(1+getTanPsi()*getTanPsi())*2.*pow(10,-3)*
				getBeta()/6.*cos(convertDeg(getAngle()))*deltaTheta/sin(convertDeg(getAngle())+deltaTheta);
	}
	else
	{
		return 0.;
	}
}

/**
 * @brief Getter for delta r
 *
 * @param rNum random number class
 *
 * @return delta r
 */
double BurgerPotgieter::getDeltaR(Random &rNum) const {
	double rr_num = rNum.getNumber();

	return (- getV() + getDriftR() + getDriftSheetR()
						+ getKappaRR()/getR()*(4.+3.*getTanPsi()*getTanPsi())/(1.+getTanPsi()*getTanPsi())
						+ getKappaParallel()/getR()*2.*getTanPsi()*getTanPsi()/pow(1+getTanPsi()*getTanPsi(),2)*(perPart-1.))*dt
						+ sqrt(2.*getKappaRR()*dt)*rr_num;
}

/**
 * @brief Getter for delta theta
 *
 * @param rNum random number class
 *
 * @return delta theta
 */
double BurgerPotgieter::getDeltaTheta(Random &rNum) const {
	double rr_num = rNum.getNumber();

	return 1./(getR()*getR()*sin(getTheta()))*getKappaThetaTheta()*cos(getTheta())/(1+getTanPsi()*getTanPsi())*dt
				+ 1./getR()*getDriftTheta()*dt
				+ 1./getR()*sqrt(2.*getKappaThetaTheta()*dt)*rr_num;
}

/**
 * @brief Method to calculate all quantities which are needed several times in one step
 */
void BurgerPotgieter::calculate() {
	data.tanPsi = 2.866*0.375*getR()*sin(getTheta());
	data.rigidity = sqrt(pow(getT(),2) + 2.*getT()*getMass())/abs(getCharge());
	data.beta = data.rigidity/(getT()+getMass());
	data.parallel = kappa0*sqrt(getRigidity())*getBeta()*getB0OverB();
	data.perpendicular = perPart*data.parallel;
}

/**
 * @brief Getter for local interstellar flux
 *
 * @param Z Charge of the CR species
 *
 * @return Local interstellar flux
 */
double BurgerPotgieter::getISFlux(int Z) const {
	double fluxIS = 0;

	if (Z == 1)
	{
		fluxIS = sqrt(pow(getT(),2) + 2.*getT()*getMass())/abs(getCharge())*pow(getT()+getMass(),-3.6);
	}

	return fluxIS;
}



