/*
 * density.cc
 *
 *      Author: kappl
 */

#include <iostream>
#include "density.h"
#include "constants.h"
#include "yamada.h"
#include "jokipiiKopriva.h"
#include "burgerPotgieter.h"
#include "potgieterMoraal.h"
#include "standard2D.h"
#include "custom.h"

/**
 * @brief Constructor
 *
 * @param modelInfo Model information from import.cc
 * @param sp Particle properties
 */
Density::Density(const TmodelInformation &modelInfo, const Tproperties &sp) {
	for (int i= 0; i < modelInfo.totalNumber; i++)
	{
		particles.push_back(createParticle(modelInfo, sp));
	}
}

/**
 * @brief Destructor
 */
Density::~Density() {
	for (unsigned int i = 0; i < particles.size(); i++)
	{
		delete(particles[i]);
	}
	particles.clear();
}

/**
 * @brief Getter for a specified particle
 *
 * @param i Specified particle
 *
 * @return Particle
 */
IParticle * Density::getParticle(int i) {
	return particles[i];
}

/**
 * @brief Getter for total number of particles
 *
 * @return total number of particles
 */
int Density::getTotalNumber() const {
	return particles.size();
}

/**
 * @brief Simulate trajectory for all particles from the earth to the heliospheric boundary
 *
 * Iteration over all particles, simulate routine of Particle class will be called.
 */
void Density::simulate() {
	std::cout << Stringconst::infoSimulate << particles[0]->getT() << "..." << std::endl;
	unsigned int size = particles.size();
	Random r(1);
	long steps = 0;
	for (unsigned int i= 0; i < size; i++)
	{
		steps += particles[i]->simulate(r);
	}
	std::cout << Stringconst::infoSteps << (int) 1.*steps/size << std::endl;
}

/**
* @brief Creates particle
*
* @param modelInfo Informations from import.cc
* @param sp Particle properties (mass and charge)
*
* @return Pointer to the particle
*/
IParticle * Density::createParticle(const TmodelInformation &modelInfo, const Tproperties &sp) const {
	if (modelInfo.model == Stringconst::modelYamada)
	{
		return new Yamada(modelInfo, sp);
	}
	else if (modelInfo.model == Stringconst::modelJokipii)
	{
		return new JokipiiKopriva(modelInfo, sp);
	}
	else if (modelInfo.model == Stringconst::modelBurger)
	{
		return new BurgerPotgieter(modelInfo, sp);
	}
	else if (modelInfo.model == Stringconst::modelPotgieter)
	{
		return new PotgieterMoraal(modelInfo, sp);
	}
	else if (modelInfo.model == Stringconst::modelStandard2D)
	{
		return new Standard2D(modelInfo, sp);
	}
	else
	{
		return new Custom(modelInfo, sp);
	}
	return NULL;
}
