/*
 * fileHelper.cc
 *
 *      Author: kappl
 */

#include "fileHelper.h"

/**
 * @brief Constructor
 *
 */
FileHelper::FileHelper() {
}

/**
 * @brief Destructor
 */
FileHelper::~FileHelper() {
}

/**
 * @brief Read input file and store result in a map
 *
 * @param input File name of the input file
 * @param first Number of the first column to be imported
 * @param second Number of the second column to be imported
 *
 * @return Map of the columns from the input file
 */
template <typename T1, typename T2> std::map<T1,T2> FileHelper::readFile(std::string input, int first, int second)
{
	std::string line;
	std::string dummy;
	std::ifstream file(input);

	std::map<T1, T2> data;

	if (file.is_open())
	{
		while (std::getline(file, line))
	    {
			std::stringstream s(line);
			T1 key;
			T2 value;
			// Skip all columns before first
			for (int i = 1; i < first; i++)
			{
				s >> dummy;
			}
			// Store this column as key
			s >> key;
			// Skip all columns before second
			for (int i = first+1; i < second; i++)
			{
				s >> dummy;
			}
			// Store this column as value
			s >> value;

			// Attach key value pair to data
			data[key] = value;
	    }
		std::cout << Stringconst::infoData << input << std::endl;
	    file.close();
	}
	else
	{
		std::cerr << Stringconst::errorOpenFile << std::endl;
	}
	return data;
}

/**
 * Explicit Instantiation
 */
template std::map<std::string, std::string> FileHelper::readFile<std::string, std::string>(std::string, int, int);
template std::map<int, double> FileHelper::readFile<int, double>(std::string, int, int);
template std::map<double, double> FileHelper::readFile<double, double>(std::string, int, int);
template std::map<std::string, double> FileHelper::readFile<std::string, double>(std::string, int, int);

/**
 * @brief Read neutron monitor data from file and put data in a map
 *
 * @param input File name for input file
 *
 * @return Map with data for each year and month
 */
std::map<std::pair<int, int>,double> FileHelper::readNeutronMonitorData(std::string input)
{
	std::string line;
	std::string dummy;
	std::ifstream file(input);

	std::map<std::pair<int, int>, double> data;

	if (file.is_open())
	{
		while (std::getline(file, line))
	    {
			std::stringstream s(line);
 			std::pair<int, int> key;
 			// Initialization necessary to avoid bad data in the case of blank lines
 			int year = 0;
			double value;
			s >> year;
			// Parse line
			for (unsigned int i = 0; i < 12; i++)
			{
				s >> value;
				key = {year, i+1};
				data[key] = value;
				//std::cout << value << std::endl;
				//std::cout << data[key] << std::endl;
			}
	    }
		std::cout << Stringconst::infoData << input << std::endl;
	    file.close();
	}
	else
	{
		std::cerr << Stringconst::errorOpenFile << std::endl;
	}
	return data;
}

/**
 * @brief Text output to a given file name
 *
 * @param output File name of the output file
 * @param ekin Vector with kinetic energy
 * @param flux Vector with CR flux
 * @param j CR species which should be in the output
 */
void FileHelper::writeText(std::string output, const std::vector<double> &ekin, const std::vector<std::vector<double>> &flux, int j) {
	std::cout << std::endl << Stringconst::infoTextOutput + output + ".dat" << std::endl;

	std::ofstream file(output + ".dat");

	if (file.is_open())
	{
		for (unsigned int i = 0; i < ekin.size(); ++i)
		{
			file << ekin[i];
			if ( j < 0)
			{
				for (unsigned int k = 0; k < flux.size(); ++k)
				{
					file << "    " << flux[k][i];
				}
				file << std::endl;
			}
			else
			{
				file << "    " << flux[j][i] << std::endl;
			}
		}
		file.close();
	}
	else
	{
		std::cerr << Stringconst::errorOpenFile << std::endl;
	}
}

/**
 * @brief Routine to open a file and to check if the format is FITS or not
 *
 * @param input File name of the input file
 *
 * @return String with the file format
 */
std::string FileHelper::checkFormat(std::string input) {
	std::string line, format;
	std::ifstream file(input);
	if (file.is_open())
	{
		getline(file,line);
	    if (line.find("SIMPLE") != std::string::npos)
		{
		   	std::cout << Stringconst::infoFITSFormat << std::endl;
			format = "FITS";
		}
		else
		{
		  	std::cout << Stringconst::infoTextFormat << std::endl;
		   	format = "Text";
		}
	    file.close();
	}
	else
	{
		std::cerr << Stringconst::errorOpenFile << std::endl;
	}
	return format;
}

