/*
 * import.cc
 *
 *      Author: kappl
 */

#include "import.h"

/**
 * @brief Constructor
 *
 * @param argc Number of given parameters
 * @param argv Parameters
 */
Import::Import(int argc, const char **argv) {
	std::cout << std::endl << Stringconst::infoVersion << std::endl << std::endl;

	if (argc < 2)
	{
		std::cerr << Stringconst::errorArgumentMissingFirst  << argv[0] << Stringconst::errorArgumentMissingSecond << std::endl;
	}

	std::string controlFileName, angleFileName, nmFileName;
	status = success;

	for (int i = 1; i < argc; ++i)
	{
		std::string arg = argv[i];
		if ((arg == "-c") || (arg == "--control"))
		{
			if (i + 1 < argc)
			{
				controlFileName = argv[i+1];
			}
			else
			{
				std::cerr << Stringconst::errorControlMissing << std::endl;
				status = error;
				return;
			}
		}
		else if ((arg == "-a") || (arg == "--angle"))
		{
			if (i + 1 < argc)
			{
				angleFileName = argv[i+1];
			}
			else
			{
				std::cerr << Stringconst::errorAngleMissing << std::endl;
				status = error;
				return;
			}
		}
		else if ((arg == "-n") || (arg == "--neutron"))
		{
			if (i + 1 < argc)
			{
				nmFileName = argv[i+1];
			}
			else
			{
				std::cerr << Stringconst::errorNMMissing << std::endl;
				status = error;
				return;
			}
		}
		else if ((arg == "-o") || (arg == "--output"))
		{
			if (i + 1 < argc)
			{
				outputFileName = argv[i+1];
			}
			else
			{
				std::cerr << Stringconst::errorOutputMissing << std::endl;
				status = error;
				return;
			}
		}
		else if ((arg == "-i") || (arg == "--input"))
		{
			if (i + 1 < argc)
			{
				inputFileName = argv[i+1];
			}
			else
			{
				std::cerr << Stringconst::errorInputMissing << std::endl;
				status = error;
				return;
			}
		}
		else if ((arg == "-v") || (arg == "--version"))
		{
			status = version;
			return;
		}
	}
	// Read control data
	FileHelper newFile;
	if (!controlFileName.empty())
	{
		controlData = newFile.readFile<std::string, std::string>(controlFileName,1,2);
	}
	// Parse input parameters
	parse(angleFileName, nmFileName, inputFileName);
	// Ensure that outputFileName is not empty
	if (outputFileName.empty())
	{
		outputFileName = "test";
	}
}

/**
 * @brief Destructor
 */
Import::~Import() {
}

/**
 * @brief Template to convert string from a map to arbitrary type T
 *
 * @param key Key of a map
 *
 * @return Converted value of the map
 */
template <typename T> T Import::convert(std::string key)
{
	T result;
	if (controlData.find(key) != controlData.end())
	{
		std::stringstream s(controlData[key]);
		if (s >> result)
		{
			s >> result;
		}
		else
		{
			std::cout << Stringconst::warningWrongTypeFirst << key << Stringconst::warningWrongTypeSecond << std::endl;
		}
	}
	return result;
}

/**
 * @brief Check if key is in controlData and return corresponding value
 *
 * @param key Key
 * @param defaultValue If key is not found return this value
 * @param messageSuccess Message if key was found
 * @param messageError Message to inform the user if key was not found
 *
 * @return Return value
 */
template <typename T> T Import::check(std::string key, T defaultValue, std::string messageSuccess, std::string messageError)
{
	T result;
	if (controlData.find(key) != controlData.end())
	{
		result = convert<T>(key);
		std::cout << messageSuccess << result << std::endl;
	}
	else
	{
		result = defaultValue;
		if (!messageError.empty())
		{
			std::cout << messageError << std::endl;
		}
	}
	return result;
}

/**
 * @brief Read particle properties from control.dat
 */
void Import::readProperties() {
	spec.mass = check<double>("mass", 0.938, Stringconst::infoMass, Stringconst::warningNoMass);
	spec.charge = check<int>("charge", 1, Stringconst::infoCharge, Stringconst::warningNoCharge);
}

/**
 * @brief Method to parse the input data and replace by default values if necessary
 *
 * @param angleFile Filename of the file with tilt angle information
 * @param nmFile Filename of the file with neutron monitor information
 * @param inputFile Filename of the LIS flux
 */
void Import::parse(std::string angleFile, std::string nmFile, std::string inputFile) {
	FileHelper newFile;
	// Get model
	if (controlData.find("model") != controlData.end())
	{
		modelInfo.model = convert<std::string>("model");
		// Input validation
		if (allowedModels.find(modelInfo.model) == allowedModels.end())
		{
			std::cout << Stringconst::warningModelNotAllowed << std::endl;
			modelInfo.model = Stringconst::modelYamada;
		}
		else
		{
			std::cout << Stringconst::infoModel << modelInfo.model << std::endl;
		}
	}
	else
	{
		modelInfo.model = Stringconst::modelYamada;
		std::cout << Stringconst::warningNoModel << std::endl;
	}

	// Get force field value
	phi = check<double>("phi", -1, Stringconst::infoForceField, Stringconst::infoNoForceField);

	// Get modNumber if provided
	modNumber = check<int>("modNumber", -1, Stringconst::infoModNumber, Stringconst::infoNoModNumber);

	// Check if input file is provided and its format
	if (!inputFile.empty())
	{
		inputFormat = newFile.checkFormat(inputFile);
		if (inputFormat == "Text")
		{
			readProperties();
		}
	}
	else
	{
		std::cout << Stringconst::infoNoFlux << std::endl;
		// No input file provided, take model data instead
		inputFormat = "Model";
		// Read particle properties from control.dat file
		readProperties();
	}

	// Check if time info is necessary and if yes, read it
	int year, month, yearEnd, monthEnd;
	if (!angleFile.empty() || !nmFile.empty())
	{
		// Determine date
		year = check<int>("year", 2015, Stringconst::infoYear, Stringconst::warningNoYear);
		month = check<int>("month", -1, Stringconst::infoMonth, Stringconst::warningNoMonth);

		// Get end date if provided
		yearEnd = check<int>("yearEnd", -1, Stringconst::infoYearEnd, "");
		monthEnd = check<int>("monthEnd", -1, Stringconst::infoMonthEnd, "");
	}

	// Check if file with tilt angle information is provided. If yes, take tilt angle from file
	if (!angleFile.empty())
	{
		// Get tilt angle model (L or R model)
		char tiltModel = check<char>("tiltModel", 'R', Stringconst::infoTiltModel, Stringconst::warningNoTiltModel);

		if (tiltModel == 'R')
		{
			angleData = newFile.readFile<int, double>(angleFile,2,5);
		}
		else if (tiltModel == 'L')
		{
			angleData = newFile.readFile<int, double>(angleFile,2,8);
		}
		else
		{
			angleData = newFile.readFile<int, double>(angleFile,2,5);
			std::cout << Stringconst::warningInvalidTiltModel << std::endl;
		}
		// Set default
		modelInfo.polarity = 1;
		modelInfo.angle = 0;

		int carrington = 0;

		// If no month is provided, average over the whole year
		if (month == -1)
		{
			modelInfo.angle = averageAngle(year, 1, 1, year, 12, 31);
			carrington = calculateCR(year, 1, 1);
		}
		else
		{
			if (monthEnd > -1 && yearEnd > -1)
			{
				modelInfo.angle = averageAngle(year, month, 1, yearEnd, monthEnd, 31);
			}
			else
			{
				modelInfo.angle = averageAngle(year, month, 1, year, month, 31);
			}
			carrington = calculateCR(year, month, 15);
		}

		std::cout << Stringconst::infoAngle << modelInfo.angle << std::endl;

		// Determine polarity from Carrington rotation
		if (carrington < 1689)
		{
			modelInfo.polarity = 1;
		}
		else if (carrington >= 1689 && carrington < 1837)
		{
			modelInfo.polarity = -1;
		}
		else if (carrington >= 1837 && carrington < 1955)
		{
			modelInfo.polarity = 1;
		}
		else if (carrington >= 1955 && carrington < 2130)
		{
			modelInfo.polarity = -1;
		}
		else
		{
			modelInfo.polarity = 1;
		}

		std::cout << Stringconst::infoPolarity << modelInfo.polarity << std::endl;
	}
	else
	{
		// Get polarity
		modelInfo.polarity = check<int>("polarity", 1, Stringconst::infoPolarityDet, Stringconst::warningNoPolarity);
		// Get tilt angle
		modelInfo.angle = check<double>("angle", 0, Stringconst::infoAngleDet, Stringconst::warningNoTiltAngle);
	}

	// Get output Format
	outputFormat = check<std::string>("outputFormat", "Text", Stringconst::infoOutput, Stringconst::warningNoOutputFormat);
	// Get kappaZero if given
	modelInfo.kappaScaling = check<double>("kappaScaling", 1., Stringconst::infoKappa, "");
	// Get BfieldScaling if given
	modelInfo.BfieldScaling = check<double>("BfieldScaling", 1., Stringconst::infoB, "");
	// Get index if given
	index = check<double>("index", 0, Stringconst::infoIndex, "");
	// Get dt if given
	modelInfo.dt = check<int>("dt", -1, Stringconst::infoDeltaTime, "");
	// Get extra if given
	extraBins = check<int>("extraBins", 5, Stringconst::infoExtraBins, "");
	// Get total number of particles which should be simulated
	modelInfo.totalNumber = check<int>("total", 1000, Stringconst::infoTotalNumber, Stringconst::infoNoTotalNumber);

	// Get neutron monitor data if standard model is selected
	if (!nmFile.empty() && modelInfo.model != Stringconst::modelYamada
				&& modelInfo.model != Stringconst::modelJokipii
				&& modelInfo.model != Stringconst::modelPotgieter
				&& modelInfo.model != Stringconst::modelBurger)
	{
		nmData = newFile.readNeutronMonitorData(nmFile);
		double nmValue = 0;

		// If no month is provided, average over the whole year
		if (month == -1)
		{
			nmValue = averageNeutronMonitorData(year, 1, year, 12);
		}
		else
		{
			if (monthEnd > -1 && yearEnd > -1)
			{
				nmValue = averageNeutronMonitorData(year, month, yearEnd, monthEnd);
			}
			else
			{
				nmValue = averageNeutronMonitorData(year, month, year, month);
			}
		}

		std::cout << Stringconst::infoNMValue << nmValue << std::endl;

		calculateKappaZero(nmValue);
	}
	else if (nmFile.empty() && modelInfo.model != Stringconst::modelYamada
			&& modelInfo.model != Stringconst::modelJokipii
			&& modelInfo.model != Stringconst::modelPotgieter
			&& modelInfo.model != Stringconst::modelBurger)
	{
		double nmValue = check<double>("nmValue", 500, Stringconst::infoNMDet, Stringconst::warningNoNMValue);
		calculateKappaZero(nmValue);
	}
}

/**
 * @brief Calculate Carrington rotation from date
 *
 * @param year
 * @param month
 * @param day
 *
 * @return Carrington rotation number
 */
int Import::calculateCR(int year, int month, int day) {
	// Calculate Julian date
	int A = floor(year/100.);
	int B = 2 - A + floor(A/4.);
	double jd = 0;
	if (month > 2)
	{
		jd = floor(365.25*(year+4716)) + floor(30.6001*(month+1)) + B + day - 1524.5;
	}
	else
	{
		jd = floor(365.25*((year-1.)+4716)) + floor(30.6001*((month+12)+1)) + B + day - 1524.5;
	}
	// Calculate Carrington rotation
	return (int) (jd - 2398140.227)/27.2752316;
}

/**
 * @brief Calculate average tilt angle from two dates
 *
 * @param startYear
 * @param startMonth
 * @param startDay
 * @param endYear
 * @param endMonth
 * @param endDay
 *
 * @return Averaged tilt angle
 */
double Import::averageAngle(int startYear, int startMonth, int startDay, int endYear, int endMonth, int endDay) {
	double average = 0;
	// Convert Julian date to Carrington rotation
	int carringtonStart = calculateCR(startYear, startMonth, startDay);
	int carringtonEnd = calculateCR(endYear, endMonth, endDay);
	// Sum and calculate average
	int counter = 0;
	for (int i = carringtonStart; i <= carringtonEnd; i++)
	{
		average += angleData[i];
		counter++;
	}
	average = (1.*average)/counter;

	return average;
}

/**
 * @brief Calculate average neutron monitor data from two dates
 *
 * @param startYear
 * @param startMonth
 * @param endYear
 * @param endMonth
 *
 * @return Averaged neutron monitor data
 */
double Import::averageNeutronMonitorData(int startYear, int startMonth, int endYear, int endMonth) {
	double average = 0;
	int counter = 0;
	for (int i = startYear; i <= endYear; i++)
	{
		if (i == startYear && i == endYear)
		{
			for (int k = startMonth; k <= endMonth; k++)
			{
				average += nmData[{i, k}];
				counter++;
			}
		}
		else if (i == startYear)
		{
			for (int k = startMonth; k <= 12; k++)
			{
				average += nmData[{i, k}];
				counter++;
			}
		}
		else if (i == endYear)
		{
			for (int k = 1; k <= endMonth; k++)
			{
				average += nmData[{i, k}];
				counter++;
			}
		}
		else
		{
			for (int k = 1; k <= 12; k++)
			{
				average += nmData[{i, k}];
				counter++;
			}
		}
	}
	average = (1.*average)/counter;

	return average;
}

/**
 * @brief Method to calculate diffusion tensor normalization from Neutron monitor data
 *
 * @param nmValue Averaged Neutron Monitor value
 */
void Import::calculateKappaZero(double nmValue) {
	modelInfo.kappaScaling = modelInfo.kappaScaling *
			(137*pow(nmValue, -1) - 0.061);

	std::cout << Stringconst::infoKappaComputation << modelInfo.kappaScaling << std::endl;
}

/**
 * @brief Getter for phi
 *
 * @return Phi
 */
double Import::getPhi() const {
	return phi;
}

/**
 * @brief Getter for outputFormat
 *
 * @return OutputFormat (Text or FITS)
 */
std::string Import::getOutputFormat() const {
	return outputFormat;
}

/**
 * @brief Getter for inputFormat
 *
 * @return InputFormat (Text or FITS)
 */
std::string Import::getInputFormat() const {
	return inputFormat;
}

/**
 * @brief Getter for outputFileName
 *
 * @return File name of the output file
 */
std::string Import::getOutputFileName() const {
	return outputFileName;
}

/**
 * @brief Getter for inputFileName
 *
 * @return File name of the input file
 */
std::string Import::getInputFileName() const {
	return inputFileName;
}

/**
 * @brief Getter for particle properties
 *
 * @return spec (Particle properties)
 */
Tproperties Import::getSpec() const {
	return spec;
}

/**
 * @brief Getter for modNumber
 *
 * @return Cosmic ray species which should be modulated
 */
int Import::getModNumber() const {
	return modNumber;
}

/**
 * @brief Getter for index
 *
 * @return spectral index for output rescaling
 */
double Import::getIndex() const {
	return index;
}

/**
 * @brief Getter for extraBins
 *
 * @return number of extra bins for quality improvement
 */
int Import::getExtraBins() const {
	return extraBins;
}

/**
 * @brief Getter for ModelInformation
 *
 * @return ModelInformation
 */
TmodelInformation Import::getModelInformation() const {
	return modelInfo;
}

/**
 * @brief Getter for ImportStatus
 *
 * @return ImportStatus
 */
TimportStatus Import::getImportStatus() const {
	return status;
}
