/*
 * jokipiiKopriva.cc
 *
 *      Author: kappl
 */

#include "jokipiiKopriva.h"
#include <cmath>
#include <chrono>
#include <iostream>

/**
 * @brief Constructor
 *
 * @param modelInfo Model information from import.cc
 * @param sp Particle properties
 */
JokipiiKopriva::JokipiiKopriva(const TmodelInformation &modelInfo, const Tproperties &sp) : Particle(modelInfo, sp) {
	if (modelInfo.dt > 0)
	{
		dt = modelInfo.dt;
	}
	else
	{
		dt = 1000;
	}
	// Equals 5*10^21 cm^2 s^-1 GV^-1
	kappa0 = modelInfo.kappaScaling*20./9.*pow(10,-5);
	perPart = 0.1;
	heliosphereBoundary = 10.;
}

/**
 * @brief Destructor
 */
JokipiiKopriva::~JokipiiKopriva() {
}

/**
 * @brief Getter for parallel part of the diffusion tensor
 *
 * @return Parallel part of the diffusion tensor
 */
double JokipiiKopriva::getKappaParallel() const {
	return data.parallel;
}

/**
 * @brief Getter for radial part of the diffusion tensor
 *
 * @return Radial part of the diffusion tensor
 */
double JokipiiKopriva::getKappaRR() const {
	return getKappaParallel()/(1.+getTanPsi()*getTanPsi())
				+ getKappaPerpendicular()*getTanPsi()*getTanPsi()/(1.+getTanPsi()*getTanPsi());
}

/**
 * @brief Getter for drift in r direction
 *
 * @return radial drift
 */
double JokipiiKopriva::getDriftR() const {
	// Define Heaviside function depending term
	int heavy = getHeaviside();

	return getPolarity()*getCharge()*9./2.*pow(10,-14)*2./3.*getRigidity()*getBeta()/(Constants::A*BfieldScaling)*getR()*cos(getTheta())/sin(getTheta())*getTanPsi()/pow(1+getTanPsi()*getTanPsi(),2)*heavy;
}

/**
 * @brief Getter for drift in theta direction
 *
 * @return angular drift
 */
double JokipiiKopriva::getDriftTheta() const {
	// Define Heaviside function depending term
	int heavy = getHeaviside();

	return -1.*getPolarity()*getCharge()*9./2.*pow(10,-14)*2./3.*getRigidity()*getBeta()/(Constants::A*BfieldScaling)*getR()*getTanPsi()*(2+getTanPsi()*getTanPsi())/pow(1+getTanPsi()*getTanPsi(),2)*heavy;
}

/**
 * @brief Getter for drift in the heliospheric current sheet (HCS) in radial direction
 *
 * @return HCS drift
 */
double JokipiiKopriva::getDriftSheetR() const {
	// Distance to the current sheet
	double d = fabs(getR()*cos(getTheta()));

	// There is only a drift effect close to the current sheet
	if (d < 2.*getLarmorRadius())
	{
		// Normalization 2 * 10^-3 to get speed in AU/s
		return -1.*getPolarity()*getCharge()*getTanPsi()/sqrt(1+getTanPsi()*getTanPsi())*2.*pow(10,-3)*
				(0.457-0.412*d/getLarmorRadius()+0.0915*pow(d,2)/pow(getLarmorRadius(),2))*getBeta();
	}
	else
	{
		return 0;
	}
}

/**
 * @brief Getter for delta r
 *
 * @param rNum random number class
 *
 * @return delta r
 */
double JokipiiKopriva::getDeltaR(Random &rNum) const {
	double rr_num = rNum.getNumber();

	return (- getV() + getDriftR() + getDriftSheetR()
					+ 2./getR()*(getKappaRR() + getKappaParallel()*getTanPsi()*getTanPsi()/pow(1+getTanPsi()*getTanPsi(),2)*(perPart-1.)))*dt
					+ sqrt(2.*(getKappaRR())*dt)*rr_num;
}

/**
 * @brief Getter for delta theta
 *
 * @param rNum random number class
 *
 * @return delta theta
 */
double JokipiiKopriva::getDeltaTheta(Random &rNum) const {
	double rr_num = rNum.getNumber();

	return 1./(getR()*getR()*sin(getTheta()))*getKappaThetaTheta()*cos(getTheta())*(dt)
					+ 1./getR()*getDriftTheta()*dt
					+ 1./getR()*sqrt(2.*getKappaThetaTheta()*dt)*rr_num;
}

/**
 * @brief Method to calculate all quantities which are needed several times in one step
 */
void JokipiiKopriva::calculate() {
	data.tanPsi = 2.866*0.375*getR()*sin(getTheta());
	data.rigidity = sqrt(pow(getT(),2) + 2.*getT()*getMass())/abs(getCharge());
	data.beta = data.rigidity/(getT()+getMass());
	data.parallel = kappa0*sqrt(getRigidity())*getBeta();
	data.perpendicular = perPart*data.parallel;
}

/**
 * @brief Getter for local interstellar flux
 *
 * @param Z Charge of the CR species
 *
 * @return Local interstellar flux
 */
double JokipiiKopriva::getISFlux(int Z) const {
	double fluxIS = 0;

	if (Z == 1)
	{
		fluxIS = sqrt(pow(getT(),2) + 2.*getT()*getMass())/abs(getCharge())*pow(getMass()*getMass()+(pow(getT(),2) + 2.*getT()*getMass())/pow(abs(getCharge()),2),-1.8);
	}

	return fluxIS;
}


