/*
 * particle.cc
 *
 *      Author: kappl
 */

#include "particle.h"
#include "random.h"
#include "constants.h"
#include "stringconst.h"
#include <cmath>
#include <iostream>
#include <fstream>
#include <chrono>

/**
 * @brief Constructor
 *
 * @param modelInfo Model information from import.cc
 * @param sp CR properties (mass and charge)
 */
Particle::Particle(const TmodelInformation &modelInfo, const Tproperties &sp) {
	r = Constants::positionEarth;
	theta = 0.5*Constants::pi;
	Random rNum(1);
	phi = rNum.getRandZero2Pi();
	time = 0;

	spec.T = sp.T;
	spec.mass = sp.mass;
	spec.charge = sp.charge;

	dt = 0;
	heliosphereBoundary = 0;
	perPart = 0;
	kappa0 = modelInfo.kappaScaling;
	BfieldScaling = modelInfo.BfieldScaling;
	polarity = modelInfo.polarity;
	angle = modelInfo.angle;

	data.alphaH = 0;
	data.beta = 0;
	data.parallel = 0;
	data.perpendicular = 0;
	data.rigidity = 0;
	data.tanPsi = 0;
}

/**
 * @brief Destructor
 */
Particle::~Particle() {
}

/**
 * @brief Getter for r
 *
 * @return r
 */
double Particle::getR() const {
	return r;
}

/**
 * @brief Getter for kinetic energy T
 *
 * @return T
 */
double Particle::getT() const {
	return spec.T;
}

/**
 * @brief Getter for angle theta
 *
 * @return theta
 */
double Particle::getTheta() const {
	return theta;
}

/**
 * @brief Getter for angle phi
 *
 * @return phi
 */
double Particle::getPhi() const {
	return phi;
}

/**
 * @brief Getter for time
 *
 * @return time
 */
double Particle::getTime() const {
	return time;
}

/**
 * @brief Getter for mass
 *
 * @return mass
 */
double Particle::getMass() const {
	return spec.mass;
}

/**
 * @brief Getter for charge
 *
 * @return charge
 */
double Particle::getCharge() const {
	return spec.charge;
}

/**
 * @brief Simulate one iteration
 *
 * @param rNum random number class
 */
void Particle::step(Random &rNum) {
	// Compute tan(psi) etc.
	calculate();

	// Variables to store variables temporarily
	double rTemp = getR();
	double timeTemp = getTime();
	double TTemp = getT();
	double thetaTemp = getTheta();
	double phiTemp = getPhi();

	// Change pseudo-particle values
	rTemp += getDeltaR(rNum);
	timeTemp += getDeltaTime();
	TTemp += getDeltaT();

	thetaTemp += getDeltaTheta(rNum);
	// Take care that theta is always between 0 and pi (see Strauss et al.)
	if (thetaTemp < 0)
	{
		thetaTemp = fabs(thetaTemp);
	}

	if (thetaTemp > 2.*Constants::pi)
	{
		thetaTemp = thetaTemp - 2.*Constants::pi;
	}
	else if (thetaTemp > Constants::pi)
	{
		thetaTemp = 2.*Constants::pi - thetaTemp;
	}

	phiTemp += getDeltaPhi();
	// Take care that phi is always between 0 and 2*pi (see Strauss et al.)
	if (phiTemp < 0)
	{
		phiTemp += 2*Constants::pi;
	}
	if (phiTemp > 2*Constants::pi)
	{
		phiTemp -= 2*Constants::pi;
	}

	// Define radial boundary condition for r = 0
	if (rTemp < 0)
	{
		rTemp = -1.*rTemp;
	}

	r = rTemp;
	time = timeTemp;
	spec.T = TTemp;
	theta = thetaTemp;
	phi = phiTemp;
}

/**
 * @brief Simulate trajectory from the earth to the heliospheric boundary
 *
 * @param rNum random number class
 *
 * @return Number of needed steps
 */
long Particle::simulate(Random &rNum) {
	long counter = 0;
	// Simulate only inside the heliosphere and below the time limit
	while(r < getHeliosphereBoundary() && time < Constants::timeLimit)
	{
		step(rNum);
		counter++;
	}
	return counter;
}

/**
 * @brief Getter for delta T
 *
 * @return delta T
 */
double Particle::getDeltaT() const {
	return 2.*getV()/(3.*getR())*(pow(getT(),2) + 2.*getT()*getMass())/(getT() + getMass())*dt;
}

/**
 * @brief Getter for delta theta
 *
 * @param rNum random number class
 *
 * @return delta theta
 */
double Particle::getDeltaTheta(Random &rNum) const {
	return 0;
}

/**
 * @brief Getter for delta phi
 *
 * @return delta phi
 */
double Particle::getDeltaPhi() const {
	return 0;
}

/**
 * @brief Getter for delta t
 *
 * @return delta t
 */
long Particle::getDeltaTime() const {
	return dt;
}

/**
 * @brief Getter for solar wind in AU*s^-1
 *
 * @return solar wind
 */
double Particle::getV() const {
	// 400 km*s^-1 converted to AU*s^-1
	return 8./3.*pow(10,-6);
}

/**
 * @brief Getter for rigidity^2 in GV^2
 *
 * @return rigidity^2
 */
double Particle::getP2() const {
	return (pow(getT(),2) + 2.*getT()*getMass())*pow(getCharge(),-2);
}

/**
 * @brief Getter for the boundary of the heliosphere in AU
 *
 * @return Heliosphere boundary
 */
double Particle::getHeliosphereBoundary() const {
	return heliosphereBoundary;
}

/**
 * @brief Method to calculate all quantities which are needed several times in one step
 */
void Particle::calculate() {
	data.rigidity = sqrt(pow(getT(),2) + 2.*getT()*getMass())/abs(getCharge());
	data.beta = getRigidity()/(getT()+getMass());
}

/**
 * @brief Getter for rigidity in GV
 *
 * @return rigidity
 */
double Particle::getRigidity() const {
	return data.rigidity;
}

/**
 * @brief Getter for beta (particle speed)
 *
 * @return beta
 */
double Particle::getBeta() const {
	return data.beta;
}

/**
 * @brief Getter for solar polarity
 *
 * @return polarity
 */
double Particle::getPolarity() const {
	return polarity;
}

/**
 * @brief Getter for tilt angle
 *
 * @return tilt angle
 */
double Particle::getAngle() const {
	return angle;
}

/**
 * @brief Getter for tan(psi)
 *
 * @return tan(psi)
 */
double Particle::getTanPsi() const {
	return data.tanPsi;
}

/**
 * @brief Conversion method from deg to rad
 *
 * @return angle in rad
 */
double Particle::convertDeg(double deg) const {
	return deg/180.*Constants::pi;
}

/**
 * @brief Getter for B0/B (needed in many models for the diffusion tensor)
 *
 * @return B0/B
 */
double Particle::getB0OverB() const {
	return 5./3.4*pow(getR(),2)*pow(1+pow(getTanPsi(),2),-0.5);
}

/**
 * @brief Getter for Larmor radius
 *
 * @return Larmor radius
 */
double Particle::getLarmorRadius() const {
	// Convert from Tesla to GeV/AU with the conversion factor 9/4 * 10^-11
	return 9./4.*pow(10,-11)*getRigidity()/getB();
}

/**
 * @brief Getter for absolute value of B field
 *
 * @return B field
 */
double Particle::getB() const {
	return BfieldScaling*Constants::A*pow(getR(),-2)*sqrt(1+pow(getTanPsi(),2));
}

/**
 * @brief Getter for Heaviside function
 *
 * @return Value of Heaviside function
 */
int Particle::getHeaviside() const {
	if (getTheta() > 0.5*Constants::pi && getTheta() < Constants::pi)
	{
		return -1;
	}
	else
	{
		return 1;
	}
}

/**
 * @brief Getter for perpendicular part of the diffusion tensor
 *
 * @return Perpendicular part of the diffusion tensor
 */
double Particle::getKappaPerpendicular() const {
	return data.perpendicular;
}

/**
 * @brief Getter for angular part of the diffusion tensor
 *
 * @return Angular part of the diffusion tensor
 */
double Particle::getKappaThetaTheta() const {
	return getKappaPerpendicular();
}

/**
 * @brief Getter for local interstellar flux
 *
 * @param check Parameter if warning should be displayed
 *
 * @return Local interstellar flux
 */
double Particle::getISFlux(bool check) const {
	double fluxIS = 0;

	// Check CR species
	if (getMass() > 0.93 && getMass() < 1.1 && getCharge() == 1)
	{
		fluxIS = getISFlux(check, 1);
	}
	else if (getMass() > 2.5 && getMass() < 4.5 && getCharge() == 2)
	{
		fluxIS = getISFlux(check, 2);
	}
	else
	{
		if (check)
		{
			std::cout << Stringconst::warningNoISFlux << std::endl;
		}
	}
	return fluxIS;
}

/**
 * @brief Getter for local interstellar flux
 *
 * @param Z Charge of the CR species
 *
 * @return Local interstellar flux
 */
double Particle::getISFlux(int Z) const {
	return 0;
}

/**
 * @brief Getter for local interstellar flux
 *
 * @param check Parameter if warning should be displayed
 * @param Z Charge of the CR species
 *
 * @return Local interstellar flux
 */
double Particle::getISFlux(bool check, int Z) const {
	if (check && getISFlux(Z) < 0.000001)
	{
		std::cout << Stringconst::warningNoISFlux << std::endl;
		return getISFlux(Z);
	}
	else
	{
		return getISFlux(Z);
	}
}
