/*
 * potgieterMoraal.cc
 *
 *      Author: kappl
 */

#include "potgieterMoraal.h"
#include <cmath>
#include <chrono>
#include <iostream>

/**
 * @brief Constructor
 *
 * @param modelInfo Model information from import.cc
 * @param sp Particle properties
 */
PotgieterMoraal::PotgieterMoraal(const TmodelInformation &modelInfo, const Tproperties &sp) : Particle(modelInfo, sp) {
	if (modelInfo.dt > 0)
	{
		dt = modelInfo.dt;
	}
	else
	{
		dt = 2000;
	}
	// Equals 0.6*10^22 cm^2 s^-1 GV^-1
	kappa0 = modelInfo.kappaScaling*0.6*4./9.*pow(10,-4);
	perPart = 2.0/0.6*0.01;
	heliosphereBoundary = 50.;

	double thetaonehalf = 85./90.*Constants::pi/2.;
	data.alphaH = acos(Constants::pi/(2.*thetaonehalf)-1.);
}

/**
 * @brief Destructor
 */
PotgieterMoraal::~PotgieterMoraal() {
}

/**
 * @brief Getter for parallel part of the diffusion tensor
 *
 * @return Parallel part of the diffusion tensor
 */
double PotgieterMoraal::getKappaParallel() const {
	return data.parallel;
}

/**
 * @brief Getter for radial part of the diffusion tensor
 *
 * @return Radial part of the diffusion tensor
 */
double PotgieterMoraal::getKappaRR() const {
	return getKappaParallel()/(1.+getTanPsi()*getTanPsi())
				+ getKappaPerpendicular()*getTanPsi()*getTanPsi()/(1.+getTanPsi()*getTanPsi());
}

/**
 * @brief Getter for drift in r direction
 *
 * @return radial drift
 */
double PotgieterMoraal::getDriftR() const {
	return getPolarity()*getCharge()*9./2.*pow(10,-14)*2./3.*getRigidity()*getBeta()/(Constants::A*BfieldScaling)*getR()*cos(getTheta())/sin(getTheta())*getTanPsi()/pow(1+getTanPsi()*getTanPsi(),2)*getf();
}

/**
 * @brief Getter for drift in theta direction
 *
 * @return angular drift
 */
double PotgieterMoraal::getDriftTheta() const {
	return -getPolarity()*getCharge()*9./2.*pow(10,-14)*2./3.*getRigidity()*getBeta()/(Constants::A*BfieldScaling)*getR()*getTanPsi()*(2+getTanPsi()*getTanPsi())/pow(1+getTanPsi()*getTanPsi(),2)*getf();
}

/**
 * @brief Getter for drift in the heliospheric current sheet (HCS) in radial direction
 *
 * @return HCS drift
 */
double PotgieterMoraal::getDriftSheetR() const {
	return getPolarity()*getCharge()*9./2.*pow(10,-14)*1./3.*getRigidity()*getBeta()/(Constants::A*BfieldScaling)*getR()*getTanPsi()/(1+getTanPsi()*getTanPsi())*getfprime();
}

/**
 * @brief Getter for delta r
 *
 * @param rNum random number class
 *
 * @return delta r
 */
double PotgieterMoraal::getDeltaR(Random &rNum) const {
	double rr_num = rNum.getNumber();

	return (- getV() + getDriftR() + getDriftSheetR()
									+ 2./getR()*(getKappaRR() + (getKappaPerpendicular() - getKappaParallel())*getTanPsi()*getTanPsi()/pow(1.+getTanPsi()*getTanPsi(),2)
									+ 1.*getKappaParallel()*getR()*getR()/(1.+getR()*getR())*1./(1.+getTanPsi()*getTanPsi())
									+ 1./2.*getKappaPerpendicular()*(2.+getTanPsi()*getTanPsi())*getTanPsi()*getTanPsi()/pow(1.+getTanPsi()*getTanPsi(),2)))*(dt)
									+ sqrt(2.*(getKappaRR())*dt)*rr_num;
}

/**
 * @brief Getter for delta theta
 *
 * @param rNum random number class
 *
 * @return delta theta
 */
double PotgieterMoraal::getDeltaTheta(Random &rNum) const {
	double rr_num = rNum.getNumber();

	return 1./(getR()*getR()*sin(getTheta()))*getKappaThetaTheta()*cos(getTheta())/(1+getTanPsi()*getTanPsi())*dt
							+ 1./getR()*getDriftTheta()*dt
							+ 1./getR()*sqrt(2.*getKappaThetaTheta()*dt)*rr_num;
}

/**
 * @brief Getter for local interstellar flux
 *
 * @param Z Charge of the CR species
 *
 * @return Local interstellar flux
 */
double PotgieterMoraal::getISFlux(int Z) const {
	double fluxIS = 0;

	if (Z == 1)
	{
		fluxIS = sqrt(pow(getT(),2) + 2.*getT()*getMass())/abs(getCharge())*pow(getMass()*getMass()+(pow(getT(),2) + 2.*getT()*getMass())/abs(getCharge()),-1.8);
	}

	return fluxIS;
}

/**
 * @brief Getter for f
 *
 * @return f
 */
double PotgieterMoraal::getf() const {
	return 1./getalphaH()*atan((1.-2.*getTheta()/Constants::pi)*tan(getalphaH()));
}

/**
 * @brief Getter for f' (derivative of f)
 *
 * @return f'
 */
double PotgieterMoraal::getfprime() const {
	return -2.*Constants::pi/getalphaH()*tan(getalphaH())/(Constants::pi*Constants::pi+pow((Constants::pi-2.*getTheta()),2)*pow(tan(getalphaH()),2));
}

/**
 * @brief Getter for alpha_HCS
 *
 * @return alpha_HCS
 */
double PotgieterMoraal::getalphaH() const {
    return data.alphaH;
}

/**
 * @brief Method to calculate all quantities which are needed several times in one step
 */
void PotgieterMoraal::calculate() {
	data.tanPsi = 2.866*0.375*getR()*sin(getTheta());
	data.rigidity = sqrt(pow(getT(),2) + 2.*getT()*getMass())/abs(getCharge());
	data.beta = data.rigidity/(getT()+getMass());

	double rigidity = getRigidity();

	if (rigidity < 0.4)
	{
		rigidity = 0.4;
	}

	data.parallel = kappa0*rigidity*getBeta()*(1. + getR()*getR());
	data.perpendicular = perPart*kappa0*getRigidity()*getBeta()*getB0OverB();
}
