/*
 * spectrum.cc
 *
 *      Author: kappl
 */

#include "spectrum.h"
#include "constants.h"
#include <iostream>
#include <fstream>
#include <sstream>
#include <vector>
#include <string>

/**
 * @brief Constructor
 *
 * @param sp List (std::vector) of particle properties
 * @param ekinInput std::vector of the spacing of the kinetic energy for the distribution
 */
Spectrum::Spectrum(const std::vector<Tproperties> &sp, const std::vector<double> &ekinInput) {
	spec = sp;
	ekin = ekinInput;
	std::vector<double> modFlux;

	for (unsigned int i = 0; i < ekin.size(); i++)
	{
		modFlux.push_back(0);
	}

	for (unsigned int j = 0; j < spec.size(); j++)
	{
		flux.push_back(modFlux);
	}
}

/**
 * @brief Constructor to read data from FITS or Text file
 *
 * @param modelInfo Model information from import.cc
 * @param input File name of the FITS or Text file
 * @param format Input Format (FITS, Text or Model)
 * @param sp Properties of CR species
 */
Spectrum::Spectrum(const TmodelInformation &modelInfo, std::string input, std::string format, Tproperties sp) {
	if (format == "FITS")
	{
	   	readFITS(input);
	}
	else if (format == "Text")
	{
		readText(input, sp);
	}
	else
	{
		std::vector<double> preFlux;
		// Check if flux for CR species is available
		sp.T = 10;
		Density den(modelInfo, sp);
		den.getParticle(1)->getISFlux(true);

		for (unsigned int i = 0; i < 30; i++)
		{
			ekin.push_back(0.01*pow(1.5,i));
			sp.T = ekin[i];
			TmodelInformation mod = modelInfo;
			mod.totalNumber = 1;
			Density den(modelInfo, sp);
			preFlux.push_back(den.getParticle(1)->getISFlux(false));
		}

		spec.push_back(sp);
		flux.push_back(preFlux);
	}
}

/**
 * @brief Destructor
 */
Spectrum::~Spectrum() {
}

/**
 * @brief Read FITS file
 *
 * @param input File name
 */
void Spectrum::readFITS(std::string input) {
	// Read FITS file
	CCfits::FITS* pInfile = new CCfits::FITS(input,CCfits::Read,true);

	// Distinguish if input is from DRAGON or from GALPROP
	if (pInfile->extension().size() > 0)
	{
		std::cout << Stringconst::infoDRAGON << std::endl;
		readDRAGON(pInfile);
	}
	else
	{
		std::cout << Stringconst::infoGALPROP << std::endl;
		readGALPROP(pInfile);
	}

	delete(pInfile);
}

/**
 * @brief Read input from GALPROP generated FITS file
 *
 * @param *pInfile Pointer to FITS object
 */
void Spectrum::readGALPROP(CCfits::FITS *pInfile) {
	// Extract primary HDU to get Ekin
	CCfits::PHDU& phdu = pInfile->pHDU();
	phdu.readAllKeys();

	double ekinfac, ekmin;
	int dime = phdu.axis(2);

	phdu.keyWord("CDELT3").value(ekinfac);
	phdu.keyWord("CRVAL3").value(ekmin);

	for (int i = 0; i < dime; i++)
	{
		// GALPROP documentation explains the calculation of the kinetic energy E in MeV as
		// log10(E/MeV) = CRVAL3 + k * CDELT3
		// which translates to
		// E in GeV = CRVAL3/100. + 10^(k * CDELT3)
		ekin.push_back(ekmin/100.*pow(10,i*ekinfac));
	}

	// Determine the number of different CR species
	unsigned int species = phdu.axis(3);

	std::cout << Stringconst::infoSpeciesFirst << species << Stringconst::infoSpeciesSecond << std::endl;

	// Read CR species
	for (unsigned int j=0; j < species; j++)
	{
		// Define CR species and fill properties
		int A, Z;

		std::stringstream conv;
		conv << std::setfill('0') << std::setw(3) << j+1;
		std::string Znumber = "NUCZ" + conv.str();
		std::string Anumber = "NUCA" + conv.str();
		phdu.keyWord(Znumber).value(Z);
		phdu.keyWord(Anumber).value(A);

		spec.push_back(checkSpecies(Z,A));
		std::vector<double> fitsFlux;

		// Read flux
		double rstart, rstep;
		// GALPROP documentation explains the calculation of the radial distance R in kpc as
		// R (kpc) = CRVAL1 + i * CDELT1
		// We use R = 8.5 kpc as position of the earth in the milky way
		phdu.keyWord("CRVAL1").value(rstart);
		phdu.keyWord("CDELT1").value(rstep);
		// Calculate index for R = 8 kpc
		int in = (int) (8.-rstart)/(1.*rstep);

		// FITS file starts with 1 and not with 0!
		std::vector<long> start = {in+1, 1, 1, j+1};
		std::vector<long> end = {in+1, 1, dime, j+1};
		std::vector<long> stride(4,1);

		std::valarray<double> contents1;
		std::valarray<double> contents2;

		// Read data for R = 8 kpc
		phdu.read(contents1,start,end,stride);

		// Set radial position value from 8 to 9 kpc
		start[0] = in + 2;
		end[0] = in + 2;

		// Read data for R = 9 kpc
		phdu.read(contents2,start,end,stride);

		// Calculate flux at the position of the earth (R = 8.5 kpc)
		// and rescale with Ekin^-2*100 because GALPROP uses flux in units MeV^2*cm^-2*s^-1*sr^-1*MeV
		for (unsigned int j=0; j < contents1.size(); j++)
		{
			fitsFlux.push_back(pow(ekin[j],-2)*10*(contents1[j] + (contents2[j]-contents1[j])/2.));
		}
		flux.push_back(fitsFlux);
	}

	std::cout << Stringconst::infoData << std::endl;
}

/**
 * @brief Read input from DRAGON generated FITS file
 *
 * @param *pInfile Pointer to FITS object
 */
void Spectrum::readDRAGON(CCfits::FITS *pInfile) {
	// Extract primary HDU to get Ekin
	CCfits::PHDU& phdu = pInfile->pHDU();
	phdu.readAllKeys();

	double ekinfac, ekmin;
	int dime = 0;

	phdu.keyWord("EKIN_FAC").value(ekinfac);
	phdu.keyWord("DIME").value(dime);
	phdu.keyWord("EKMIN").value(ekmin);

	for (int i = 0; i < dime; i++)
	{
		ekin.push_back(ekmin*pow(ekinfac,i));
	}

	// Determine the number of different CR species
	unsigned int species = pInfile->extension().size();

	std::cout << Stringconst::infoSpeciesFirst << species << Stringconst::infoSpeciesSecond << std::endl;

	// Read CR species
	for (unsigned int j=0; j < species; j++)
	{
		// Define CR species and fill properties
		int A, Z;

		CCfits::ExtHDU& image = pInfile->extension(j+1);

		// Define CR species and fill properties
		image.readAllKeys();
		image.keyWord("Z_").value(Z);
		image.keyWord("A").value(A);
		spec.push_back(checkSpecies(Z,A));

		std::vector<double> fitsFlux;

		// Read flux
		std::valarray<double> contents;
		image.read(contents);
		for (unsigned int i = 0; i < contents.size(); i++)
		{
			 fitsFlux.push_back(contents[i]);
		}
		flux.push_back(fitsFlux);
	}

	std::cout << Stringconst::infoData << std::endl;
}

/**
 * @brief Determine mass for given cosmic ray species
 *
 * @param Z Number of protons
 * @param Number of protons and neutrons
 *
 * @return TProperties object of the comsic ray species
 */
Tproperties Spectrum::checkSpecies(int Z, int A) {
	Tproperties sp;
	sp.charge = Z;

	// Define Z, A key to obtain mass of the CR species
	std::pair<int,int> massKey = {Z, A};
	// Check if (A,Z) pair was found, if not put proton mass and output warning
	if (massDict[massKey] > 0)
	{
		sp.mass = massDict[massKey];
	}
	else
	{
		std::cout << Stringconst::warningNoMassFound << std::endl;
		sp.mass = massDict[{1,1}];
	}

	return sp;
}

/**
 * @brief Read input from text file
 *
 * @param input File name of the input file
 * @param sp Properties object of the provided cosmic ray species
 */
void Spectrum::readText(std::string input, Tproperties sp) {
	FileHelper newFile;
	std::map<double, double> data = newFile.readFile<double, double>(input, 1, 2);
	std::vector<double> textFlux;
	for(std::map<double, double>::iterator it = data.begin(); it != data.end(); ++it)
	{
	   	ekin.push_back(it->first);
	   	textFlux.push_back(it->second);
	}
	flux.push_back(textFlux);
	spec.push_back(sp);
}

/**
 * @brief Output to file
 *
 * @param inputFormat Input Format
 * @param outputFormat Output Format
 * @param in Input file name
 * @param out Output file name
 * @param m modNumber for text output
 * @param index spectral index for output rescaling
 */
void Spectrum::output(std::string inputFormat, std::string outputFormat, std::string in, std::string out, int m, double index) {
	rescaleFlux(index);
	if (outputFormat == "Text")
	{
		output(out, m);
	}
	else if (outputFormat == "FITS" && inputFormat == "FITS")
	{
		output(in, out);
	}
	else if (outputFormat == "FITS" && inputFormat != "FITS")
	{
		output(out);
	}
	else if (outputFormat == "Both" && inputFormat != "FITS")
	{
		output(out, m);
		output(out);
	}
	else
	{
		output(out, m);
		output(in, out);
	}
}

/**
 * @brief Text output to a given file name
 *
 * @param output File name of the output file
 * @param j CR species which should be in the output
 */
void Spectrum::output(std::string output, int j) {
	FileHelper newFile;
	newFile.writeText(output, ekin, flux, j);
}

/**
 * @brief Output to FITS file if input was from Text file or Model
 *
 * @param output Output file name
 */
void Spectrum::output(std::string output) {

	std::cout << std::endl << Stringconst::infoFITSOutput + output + ".fit" << std::endl;

	output = output + ".fit";

	// Remove existing file
	remove(output.c_str());

	long dime = ekin.size();
	long axe[] = {2, dime};
	CCfits::FITS outFile(output, -32, 2, axe);

	outFile.pHDU().addKey("Charge", spec[0].charge, "Charge of the particles");
	outFile.pHDU().addKey("Mass", spec[0].mass, "Mass of the particles");

	// Prepare kinetic energy for first and flux for second column
	std::valarray<double> contents1(ekin.data(), ekin.size());
	std::valarray<double> contents2(flux[0].data(), flux[0].size());

	// FITS file starts with 1 and not with 0!
	std::vector<long> start = {1,1};
	std::vector<long> end = {1,dime};
	std::vector<long> stride(2,1);

	// Write data
	outFile.pHDU().write(start, end, stride, contents1);
	start[0] = 2;
	end[0] = 2;
	outFile.pHDU().write(start, end, stride, contents2);
}

/**
 * @brief Output to FITS file
 *
 * @param input Input file name
 * @param output Output file name
 */
void Spectrum::output(std::string input, std::string output) {

	std::cout << std::endl << Stringconst::infoFITSOutput + output + ".fit" << std::endl;

	output = output + ".fit";

	// Remove existing file
	remove(output.c_str());
	// open the existing FITS file
	CCfits::FITS inFile(input);

	// Distinguish if input is from DRAGON or from GALPROP
	if (inFile.extension().size() > 0)
	{
		// custom constructor FITS::FITS(const string&, const FITS&) for
		// this particular task.
		CCfits::FITS outFile(output,inFile);

		// Fill FITS file with output flux for each CR species
		for (unsigned int i=0; i < inFile.extension().size(); i++)
		{
			outFile.copy(inFile.extension(i+1));
			if (i < flux.size())
			{
				std::valarray<double> contents(flux[i].data(), flux[i].size());
				outFile.extension(i+1).write(1, contents.size(), contents);
			}
		}
	}
	else
	{
		unsigned int species = inFile.pHDU().axis(3);
		int dime = inFile.pHDU().axis(2);

		long axe[] = {1, 1, dime, species};
		CCfits::FITS outFile(output, -32, 4, axe);

		// Copy header with CR species information
		inFile.pHDU().readAllKeys();
		outFile.pHDU().copyAllKeys(&inFile.pHDU());

		// Fill FITS file with output flux for each CR species
		for (unsigned int i=0; i < species; i++)
		{
			// FITS file starts with 1 and not with 0!
			std::vector<long> start = {1, 1, 1, i+1};
			std::vector<long> end = {1, 1, dime, i+1};
			std::vector<long> stride(4,1);

			if (i < flux.size())
			{
				std::valarray<double> contents(flux[i].data(), flux[i].size());
				// Write data
				outFile.pHDU().write(start, end, stride, contents);
			}
		}
	}
}
/**
 * @brief Force field routine to calculate force field approximation for a given distribution
 *
 * @param species Number of CR species which should be modulated
 * @param is Interstellar spectrum which should be modulated
 * @param phi Modulation potential parameter phi in GV
 *
 */
void Spectrum::forceField(int species, const Spectrum &is, double phi) {
	if (species > -1)
	{
		forceField(is, phi, species);
	}
	else
	{
		forceField(is, phi);
	}
}

/**
 * @brief Force field routine to calculate force field approximation for a given distribution
 *
 * @param is Interstellar spectrum which should be modulated
 * @param phi Modulation potential parameter phi in GV
 * @param species Number of CR species which should be modulated
 *
 */
void Spectrum::forceField(const Spectrum &is, double phi, int species) {
	for (unsigned int i = 0; i < ekin.size(); i++)
	{
		unsigned int bin = i;

		for (unsigned int j = i; j < ekin.size(); j++)
		{
			if (j == 0)
			{
				if ((ekin[i] + phi > ekin[j]) &&
										(ekin[i] + phi <= ekin[j]+(ekin[j+1]-ekin[j])/2.))
				{
					bin = j;
				}
			}
			else if (j == ekin.size()-1)
			{
				if ((ekin[i] + phi > ekin[j]-(ekin[j]-ekin[j-1])/2.) &&
										(ekin[i] + phi <= ekin[j]))
				{
					bin = j;
				}
			}
			else
			{
				if ((ekin[i] + phi > ekin[j]-(ekin[j]-ekin[j-1])/2.) &&
						(ekin[i] + phi <= ekin[j]+(ekin[j+1]-ekin[j])/2.))
				{
					bin = j;
				}
			}
		}

		Tproperties specPhi = is.spec[species];
		specPhi.T = ekin[i] + phi;
		spec[species].T = ekin[i];

		double distance = ekin[bin]-(ekin[i]+phi);
		double correction = 0;
		if (distance < 0 && bin < ekin.size()-1)
		{
			correction = -(is.flux[species][bin+1]-is.flux[species][bin])/(ekin[bin+1]-ekin[bin])*distance;
		}
		else
		{
			correction = -(is.flux[species][bin]-is.flux[species][bin-1])/(ekin[bin]-ekin[bin-1])*distance;
		}

		flux[species][i] = getP2(spec[species])/getP2(specPhi)*(is.flux[species][bin] + correction);
	}
}

/**
 * @brief Force field routine to calculate force field approximation for a given distribution
 *
 * @param is Interstellar spectrum which should be modulated
 * @param phi Modulation potential parameter phi in GV
 *
 */
void Spectrum::forceField(const Spectrum &is, double phi) {
	for (unsigned int k = 0; k < spec.size(); k++)
	{
		forceField(is, phi, k);
	}
}

/**
 * @brief Rescale the distribution to obtain the flux j = p^2 * f
 *
 * @param species CR species which should be rescaled
 */
void Spectrum::rescaleWithP2(int species) {
	for (unsigned int j = 0; j < flux[species].size(); j++)
	{
		Tproperties inp = spec[species];
		inp.T = ekin[j];
		flux[species][j] = getP2(inp)*flux[species][j];
	}
}

/**
 * @brief Modulate an input distribution with the result from the simulation, given as a vector of distributions
 *
 * @param is Interstellar distribution which should be modulated
 * @param weight Vector of distributions which has been obtained from a density
 * @param species Number of CR species which should be modulated
 */
void Spectrum::modulate(const Spectrum &is, const std::vector<Distribution> &weight, int species) {

	for (unsigned int i=0; i < weight.size(); i++)
	{
		for (unsigned int j = 0; j < ekin.size(); j++)
		{
			Tproperties inp = is.spec[species];
			inp.T = ekin[j];
			if (inp.T > 0)
			{
				flux[species][i] += 1./getP2(inp)*is.flux[species][j]*weight[i].getFlux(j);
			}
		}
	}

	rescaleWithP2(species);
}
/**
 * @brief Modulate an input distribution with the result from the simulation, given as a vector of distributions
 *
 * @param species Number of CR species which should be modulated
 * @param modelInfo Model information from import.cc
 * @param is LIS spectrum
 *
 */
void Spectrum::modulate(int species, const TmodelInformation &modelInfo, const Spectrum &is) {
	if (species > -1)
	{
		modulate(modelInfo, is, species);
	}
	else
	{
		modulate(modelInfo, is);
	}
}
/**
 * @brief Modulate an input distribution with the result from the simulation, given as a vector of distributions
 *
 * @param modelInfo Model information from import.cc
 * @param is LIS spectrum
 * @param species Number of CR species which should be modulated
 *
 */
void Spectrum::modulate(const TmodelInformation &modelInfo, const Spectrum &is, int species) {
	std::vector<Distribution> weight;

	if (species + 1 > (int) spec.size())
	{
		std::cerr << Stringconst::errorModNumber << std::endl;
		return;
	}

	std::cout << std::endl << Stringconst::infoModulateFirst << is.spec[species].mass <<
			Stringconst::infoModulateSecond << is.spec[species].charge << std::endl << std::endl;

	// Simulate particles for a specific energy and put the result into the list
	for (unsigned int i = 0; i < ekin.size(); i++)
	{
		Tproperties inp = is.getSpec()[species];
		inp.T = ekin[i];
		Density den(modelInfo, inp);
		den.simulate();
		Distribution pp(den, ekin);
		weight.push_back(pp);
	}

	modulate(is, weight, species);
}

/**
 * @brief Modulate an input distribution with the result from the simulation, given as a vector of distributions
 *
 * @param modelInfo Model information from import.cc
 * @param is LIS spectrum
 *
 */
void Spectrum::modulate(const TmodelInformation &modelInfo, const Spectrum &is) {
	for (unsigned int j = 0; j < spec.size(); j++)
	{
		modulate(modelInfo, is, j);
	}
}

/**
 * @brief Function to improve the data quality
 *
 * @param extra Number of extra data points
 */
void Spectrum::quality(int extra) {
	int counter = 0;
	std::vector<double> newEkin;
	std::vector<double> newFlux;
	// Check ekin spacing
	for (unsigned int i = 0; i < ekin.size(); i++)
	{
		if (ekin[i] > 0.1 && ekin[i] < 100.)
		{
			counter++;
		}
	}
	// If ekin spacing is too low, add computed "data" points
	if (counter < 75)
	{
		for (unsigned int k = 0; k < flux.size(); k++)
		{
			for (unsigned int i = 0; i < ekin.size(); i++)
			{
				if (k == 0)
				{
					newEkin.push_back(ekin[i]);
				}
				newFlux.push_back(flux[k][i]);

				if (ekin[i] > 0.1 && ekin[i] < 100.)
				{
					double dist = (ekin[i+1]-ekin[i])/(extra +1);
					for (int j = 1; j < extra + 1; j++)
					{
						if (k == 0)
						{
							newEkin.push_back(ekin[i] + dist*j);
						}
						newFlux.push_back(flux[k][i] + (flux[k][i+1]-flux[k][i])/(ekin[i+1]-ekin[i])*dist*j);
					}
				}
			}
			for (unsigned int i = 0; i < newFlux.size(); i++)
			{
				if (i < flux[k].size())
				{
					flux[k][i] = newFlux[i];
				}
				else
				{
					flux[k].push_back(newFlux[i]);
				}
			}
			newFlux.clear();
		}
		ekin = newEkin;
	}
}

/**
 * @brief Reduce distribution to data points provided by origEkin
 *
 * @param origEkin New ekin spacing
 */
void Spectrum::reduceData(const std::vector<double> &origEkin) {
	std::vector<double> newEkin;
	std::vector<double> newFlux;
	for (unsigned int k = 0; k < flux.size(); k++)
	{
		for (unsigned int i = 0; i < origEkin.size(); i++)
		{
			for (unsigned int j = 0; j < ekin.size(); j++)
			{
				if (ekin[j] > origEkin[i] - 0.0001 && ekin[j] < origEkin[i] + 0.0001)
				{
					if (k == 0)
					{
						newEkin.push_back(ekin[j]);
					}
					newFlux.push_back(flux[k][j]);
				}
			}
		}

		for (unsigned int i = 0; i < newEkin.size(); i++)
		{
			flux[k][i] = newFlux[i];
		}
		newFlux.clear();

		for (unsigned int i = 0; i < ekin.size()-newEkin.size(); i++)
		{
			flux[k].pop_back();
		}
	}
	ekin = newEkin;
}

/**
 * @brief Method rescales flux with E^index for output
 *
 * @param index spectral index for rescaling
 */
void Spectrum::rescaleFlux(double index) {
	for (unsigned int k = 0; k < flux.size(); k++)
	{
		for (unsigned int i = 0; i < ekin.size(); i++)
		{
			flux[k][i] = pow(ekin[i], index)*flux[k][i];
		}
	}
}

/**
 * @brief Getter for Ekin
 *
 * @return Ekin
 */
std::vector<double> Spectrum::getEkin() const {
	return ekin;
}

/**
 * @brief Getter for Particle properties
 *
 * @return Properties
 */
std::vector<Tproperties> Spectrum::getSpec() const {
	return spec;
}

/**
 * @brief Getter for Particle properties
 *
 * @param i CR species
 *
 * @return Flux for CR species i
 */
std::vector<double> Spectrum::getFlux(int i) const {
	return flux[i];
}

/**
 * @brief Calculates momentum^2
 *
 * @param sp CR properties
 *
 * @return momentum^2
 */
double Spectrum::getP2(const Tproperties &sp) const {
	return (pow(sp.T,2) + 2.*sp.T*sp.mass);
}
