/*
 * standard2D.cc
 *
 *      Author: kappl
 */

#include "standard2D.h"
#include "stringconst.h"
#include <cmath>
#include <chrono>
#include <iostream>

/**
 * @brief Constructor
 *
 * @param modelInfo Model information from import.cc
 * @param sp Particle properties
 */
Standard2D::Standard2D(const TmodelInformation &modelInfo, const Tproperties &sp) : Particle(modelInfo, sp) {
	double scaling = 1.;
	if (getPolarity()*getCharge() > 0)
	{
		scaling = 0.07;
		dt = 1000;
	}
	else
	{
		scaling = 1.;
		dt = 5000;
	}
	if (modelInfo.dt > 0)
	{
		dt = modelInfo.dt;
	}
	kappa0 = scaling*modelInfo.kappaScaling*0.5*0.0015;
	perPart = 0.02;
	heliosphereBoundary = 100.;
}

/**
 * @brief Destructor
 */
Standard2D::~Standard2D() {
}

/**
 * @brief Getter for parallel part of the diffusion tensor
 *
 * @return Parallel part of the diffusion tensor
 */
double Standard2D::getKappaParallel() const {
	return data.parallel;
}

/**
 * @brief Getter for radial part of the diffusion tensor
 *
 * @return Radial part of the diffusion tensor
 */
double Standard2D::getKappaRR() const {
	return (getKappaParallel() + getKappaPerpendicular()*pow(getTanPsi(),2))/(1.+pow(getTanPsi(),2));
}

/**
 * @brief Getter for drift in r direction
 *
 * @return radial drift
 */
double Standard2D::getDriftR() const {
	return getPolarity()*getCharge()*9./2.*pow(10,-14)*2./3.*getRigidity()*getBeta()/(Constants::A*BfieldScaling)*getR()*cos(getTheta())/sin(getTheta())*getTanPsi()/pow(1+pow(getTanPsi(),2),2)*getf();
}

/**
 * @brief Getter for drift in theta direction
 *
 * @return angular drift
 */
double Standard2D::getDriftTheta() const {
	return -getPolarity()*getCharge()*9./2.*pow(10,-14)*2./3.*getRigidity()*getBeta()/(Constants::A*BfieldScaling)*getR()*getTanPsi()*(2+pow(getTanPsi(),2))/pow(1+pow(getTanPsi(),2),2)*getf();
}

/**
 * @brief Getter for drift in the heliospheric current sheet (HCS) in radial direction
 *
 * @return HCS drift
 */
double Standard2D::getDriftSheetR() const {
	return getPolarity()*getCharge()*9./2.*pow(10,-14)*1./3.*getRigidity()*getBeta()/(Constants::A*BfieldScaling)*getR()*getTanPsi()/(1+pow(getTanPsi(),2))*getfprime();
}

/**
 * @brief Getter for delta r
 *
 * @param rNum random number class
 *
 * @return delta r
 */
double Standard2D::getDeltaR(Random &rNum) const {
	double rr_num = rNum.getNumber();
	return (- getV() + getDriftR() + getDriftSheetR()
									+ 2./getR()*(getKappaRR() + (getKappaPerpendicular() - getKappaParallel())*pow(getTanPsi(),2)/pow(1.+pow(getTanPsi(),2),2)
									+ getKappaRR()*(1+1./2.*pow(getTanPsi(),2))/(1+pow(getTanPsi(),2))))*dt
									+ sqrt(2.*(getKappaRR())*dt)*rr_num;
}

/**
 * @brief Getter for delta theta
 *
 * @param rNum random number class
 *
 * @return delta theta
 */
double Standard2D::getDeltaTheta(Random &rNum) const {
	double rr_num = rNum.getNumber();
	return 1./(getR()*getR()*sin(getTheta()))*getKappaThetaTheta()*cos(getTheta())/(1+pow(getTanPsi(),2))*dt
							+ 1./getR()*getDriftTheta()*dt
							+ 1./getR()*sqrt(2.*getKappaThetaTheta()*dt)*rr_num;
}

/**
 * @brief Getter for local interstellar flux
 *
 * @param Z Charge of the CR species
 *
 * @return Local interstellar flux
 */
double Standard2D::getISFlux(int Z) const {
	double fluxIS = 0;
	
	// Check if CR is proton and use flux from Burger et al. if correct
	if (Z == 1)
	{
		fluxIS = 1.9*pow(10,4)*pow(sqrt(pow(getT(),2) + 2*getT()*0.938),-2.78)/(1 +
		    0.4866*pow(sqrt(pow(getT(),2) + 2*getT()*0.938),-2.51));
	}
	else if (Z == 2)
	{
		fluxIS = 0;
	}
	return fluxIS;
}

/**
 * @brief Getter for f
 *
 * @return f
 */
double Standard2D::getf() const {
	double arg = (1.-2.*getTheta()/Constants::pi)*tan(getalphaH());
	return atan(arg)/getalphaH();
}

/**
 * @brief Getter for f' (derivative of f)
 *
 * @return f'
 */
double Standard2D::getfprime() const {
	double arg = (1.-2.*getTheta()/Constants::pi)*tan(getalphaH());
	return -2./(getalphaH()*Constants::pi)*tan(getalphaH())/(1.+pow(arg, 2));
}

/**
 * @brief Getter for alpha_HCS
 *
 * @return alpha_HCS
 */
double Standard2D::getalphaH() const {
    return data.alphaH;
}

/**
 * @brief Method to calculate all quantities which are needed several times in one step
 */
void Standard2D::calculate() {
	data.tanPsi = 2.866*0.375*getR()*sin(getTheta());
	data.rigidity = sqrt(pow(getT(),2) + 2.*getT()*getMass())/abs(getCharge());
	data.beta = data.rigidity/(getT()+getMass());

	double rigidity = data.rigidity;

	if (rigidity < 0.1)
	{
		rigidity = 0.1;
	}

	data.parallel = kappa0/3.*rigidity*getBeta()*getB0OverB();
	data.perpendicular = perPart*data.parallel;
	data.alphaH = acos(pow(Constants::pi/sin(convertDeg(getAngle()+2.*getLarmorRadius()/getR()))-1.,-1));
}

