/*
 * yamada.cc
 *
 *      Author: kappl
 */

#include "yamada.h"
#include <cmath>
#include <chrono>
#include <iostream>

/**
 * @brief Constructor
 *
 * @param modelInfo Model information from import.cc
 * @param sp Particle properties
 */
Yamada::Yamada(const TmodelInformation &modelInfo, const Tproperties &sp) : Particle(modelInfo, sp) {
	if (modelInfo.dt > 0)
	{
		dt = modelInfo.dt;
	}
	else
	{
		dt = 15000;
	}
	// Equals 5*10^22 cm^2 s^-1 GV^-1
	kappa0 = modelInfo.kappaScaling*20./9.*pow(10,-4);
	heliosphereBoundary = 100.;
}

/**
 * @brief Destructor
 */
Yamada::~Yamada() {
}

/**
 * @brief Getter for radial part of the diffusion tensor
 *
 * @return Radial part of the diffusion tensor
 */
double Yamada::getKappaRR() const {
	return kappa0*getBeta()*getRigidity();
}

/**
 * @brief Getter for delta r
 *
 * @param rNum random number class
 *
 * @return delta r
 */
double Yamada::getDeltaR(Random &rNum) const {
	double rr_num = rNum.getNumber();

	return (- getV() + 2.*getKappaRR()/getR())*dt + sqrt(2.*getKappaRR()*dt)*rr_num;
}

/**
 * @brief Getter for local interstellar flux
 *
 * @param Z Charge of the CR species
 *
 * @return Local interstellar flux
 */
double Yamada::getISFlux(int Z) const {
	double fluxIS = 0;

	if (Z == 1)
	{
		fluxIS = sqrt(pow(getT(),2) + 2.*getT()*getMass())/abs(getCharge())*pow(getMass()*getMass()+(pow(getT(),2) + 2.*getT()*getMass())/abs(getCharge()),-1.85);
	}

	return fluxIS;
}


